from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class MeanSubscriber(PlotSubscriber):

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):

        if not isinstance(kind, str):
            raise ValueError('MeanSubscriber only accepts 1 kind')

        title        = f'{kind}_mean'
        ylabel       = 'Mean'
        xlabel       = 'Train step'
        subscription = Subscription(self, [kind], tag=tag, subsample=subsample)
        super().__init__([subscription], message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)
        self._add_publication(f'{kind}_mean', type='DATA')

    def compute(self, message):
        '''Compute the average of a quantity. A :class:`~ikkuna.export.messages.ModuleMessage`
        with the identifier ``{kind}_mean`` is published. '''

        module, module_name  = message.key

        data = message.data
        mean = data.mean()
        self._backend.add_data(module_name, mean, message.global_step)

        kind = f'{message.kind}_mean'
        self.message_bus.publish_module_message(message.global_step,
                                                message.train_step,
                                                message.epoch, kind,
                                                message.key, mean)



import torch

from ikkuna.export.messages import get_default_bus
from ikkuna.utils import ModuleTree
from ikkuna.utils import freeze_module


class Exporter(object):
    '''Class for managing publishing of data from model code.

    An :class:`Exporter` is used in the model code by either explicitly registering modules for
    tracking with :meth:`~Exporter.add_modules()` or by calling it with newly constructed modules
    which will then be returned as-is, but be registered in the process.

    .. code-block:: python

        e = Exporter(...)
        features = nn.Sequential([
            nn.Linear(...),
            e(nn.Conv2d(...)),
            nn.ReLU()
        ])

    Modules will be tracked recursively unless specified otherwise, meaning the following is
    possible:

    .. code-block:: python

        e = Exporter(...)
        e.add_modules(extremely_complex_model)
        # e will now track all layers of extremely_complex_model

    Three further changes to the training code are necessary

        #. :meth:`~Exporter.set_model()` to have the :class:`Exporter` wire up the appropriate
           callbacks.
        #. :meth:`~Exporter.set_loss()` should be called with the loss function so that
           labels can be extracted during training if if any
           :class:`~ikkuna.export.subscriber.Subscriber`\ s rely on the ``'input_labels'`` message
        #. :meth:`~Exporter.epoch_finished()` should be called if any
           :class:`~ikkuna.export.subscriber.Subscriber`\ s rely on the ``'epoch_finished'`` message

    Attributes
    ----------
    _modules    :   dict(torch.nn.Module, ikkuna.utils.NamedModule)
                    All tracked modules
    _weight_cache   :   dict
                        Cache for keeping the previous weights for computing differences
    _bias_cache :   dict
                    see ``_weight_cache``
    _model          :   torch.nn.Module
    _train_step :   int
                    Current batch index
    _global_step    :   int
                        Global step accross all epochs
    _epoch  :   int
                Current epoch
    _is_training    :   bool
                        Flag enabling/disabling some messages during testing
    _depth  :   int
                Depth to which to traverse the module tree
    _module_filter  :   list(torch.nn.Module)
                        Set of modules to capture when calling :meth:`add_modules()`. Everything not
                        in this list is ignored
    '''

    def __init__(self, depth, module_filter=None, message_bus=get_default_bus()):
        self._modules           = {}
        self._weight_cache      = {}
        self._bias_cache        = {}
        self._model             = None
        self._epoch             = 0
        # for gradient and activation to have the same step number, we need to increase it before
        # propagation or after backpropagation. but we don't know when the backprop finishes, while
        # we do know when the forward prop starts. So we step before and thus initialize the
        # counters with -1 to effectively start at 0
        self._train_step        = -1
        self._global_step       = -1
        self._is_training       = True
        self._depth             = depth
        self._frozen            = set()
        self._module_filter     = module_filter
        self._msg_bus           = message_bus
        self._current_publish_tag = 'default'

        self._epoch_started_marker = False

    @property
    def message_bus(self):
        return self._msg_bus

    @property
    def modules(self):
        '''list(torch.nn.Module) - Modules tracked by this :class:`Exporter`'''
        return list(self._modules.keys())

    @property
    def named_modules(self):
        '''list(ikkuna.utils.NamedModule) - Named modules tracked by this :class:`Exporter`'''
        return list(self._modules.values())

    def _add_module_by_name(self, named_module):
        '''Register a module with a name attached.

        Parameters
        ----------
        named_module    :   ikkuna.utils.NamedModule
        '''
        module                = named_module.module
        self._modules[module] = named_module
        module.register_forward_hook(self.new_activations)

        # for a new module, immediately cache the weights and biases. This is necessary, because
        # weights and updates need to be published in step() as only at the end of a batch (here the
        # beginning of the next one) the updates can be computed. but since we call step before the
        # forward pass, the new_activations() method has had no chance to cache the current weights
        # before the first batch starts. so we do it here.
        if hasattr(module, 'weight'):
            self._weight_cache[module] = module.weight

        if hasattr(module, 'bias') and module.bias is not None:
            self._bias_cache[module] = module.bias

        def layer_grad_hook(module, grad_in, grad_out):
            self.new_layer_gradients(module, grad_out)

        module.register_backward_hook(layer_grad_hook)

        has_bias   = hasattr(module, 'bias') and module.bias is not None
        has_weight = hasattr(module, 'weight') and module.weight is not None
        if not has_weight and not has_bias:
            return

        # For some reason, registered tensor hooks are called twice in my setup. Maybe this means
        # that the gradient is computed twice, because the grad tensors are identical. Not sure why
        # this is so.
        # cache weight and bias gradients and only call new_parameter_gradients when both are
        # received
        grad_cache = {'weight': None, 'bias': None}

        # the hooks will check whether both weight and bias have been received and if so, trigger
        # publication. If the module has no bias, then ``None`` is published for the bias component.
        # They also check whether we grads were already published at this train step and do nothing
        # in that case.
        def weight_hook(grad):
            if grad_cache['weight']:
                raise RuntimeError(f'Already received weight gradients for {named_module.name}')
            grad_cache['weight'] = grad
            if not has_bias or grad_cache['bias'] is not None:
                self.new_parameter_gradients(module, (grad_cache['weight'], grad_cache['bias']))
                grad_cache['weight'] = grad_cache['bias'] = None

        def bias_hook(grad):
            if grad_cache['bias']:
                raise RuntimeError(f'Already received bias gradients for {named_module.name}')
            grad_cache['bias'] = grad
            if not has_bias or grad_cache['weight'] is not None:
                self.new_parameter_gradients(module, (grad_cache['weight'], grad_cache['bias']))
                grad_cache['weight'] = grad_cache['bias'] = None

        if has_bias:
            module.bias.register_hook(bias_hook)
        module.weight.register_hook(weight_hook)

    def add_modules(self, module, recursive=True):
        '''Add modules to supervise. If the module has ``weight`` and/or ``bias`` members, updates
        to those will be tracked. Ignores any module in :attr:`_module_filter`.

        Parameters
        ----------
        module  :   tuple(str, torch.nn.Module) or torch.nn.Module
        recursive   :   bool
                        Descend recursively into the module tree
        depth   :   int
                    Depth to which to traverse the tree. Modules below this level will be ignored
        Raises
        ------
        ValueError
            If ``module`` is neither a tuple, nor a (subclass of) :class:`torch.nn.Module`
        '''
        if isinstance(module, tuple):   # name already given -> use that
            name, module = module
        else:
            name = module.__class__.__name__.lower()

        if isinstance(module, torch.nn.Module):
            module_tree = ModuleTree(module, name=name, recursive=recursive, drop_name=recursive)
            for named_module in module_tree.preorder(depth=self._depth):
                module, name = named_module
                if self._module_filter is None or module.__class__ in self._module_filter:
                    print('Adding ', name)
                    self._add_module_by_name(named_module)
                else:
                    print('Skipping ', name)
        else:
            raise ValueError(f'Don\'t know how to handle {module.__class__.__name__}')

    def __call__(self, module, recursive=True):
        '''Shorthand for :meth:`~Exporter.add_modules()` which returns its input unmodified.

        Parameters
        ----------
        see :meth:`Exporter.add_modules()`

        Returns
        -------
        torch.nn.Module
            The input ``module``
        '''
        self.add_modules(module, recursive)
        return module

    def train(self, train=True):
        '''Switch to training mode. This will ensure all data is published.'''
        self._is_training = train

    def test(self, test=True):
        '''Switch to testing mode. This will turn off all publishing.'''
        self.train(not test)

    def new_loss(self, loss):
        '''Callback for publishing current training loss.'''
        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'loss', loss,
                                              tag=self._current_publish_tag)

    def new_input_data(self, *args):
        '''Callback for new training input to the network.

        Parameters
        ----------
        *args   :   tuple
                    Network inputs
        '''
        if len(args) == 1:
            input_data = args[0]
        else:
            input_data = args

        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'input_data', input_data,
                                              tag=self._current_publish_tag)

    def new_output_and_labels(self, network_output, labels):
        '''Callback for final network output.

        Parameters
        ----------
        data    :   torch.Tensor
                    The final layer's output
        '''
        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'network_output', network_output,
                                              tag=self._current_publish_tag)
        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'input_labels', labels,
                                              tag=self._current_publish_tag)

    def new_activations(self, module, in_, out_):
        '''Callback for newly arriving activations. Registered as a hook to the tracked modules.
        Will trigger export of all new activation and weight/bias data.

        Parameters
        ----------
        module  :   torch.nn.Module
        in_ :   torch.Tensor
                Dunno what this is
        out_    :   torch.Tensor
                    The new activations
        '''
        if not self._epoch_started_marker:
            self._msg_bus.publish_network_message(self._global_step, self._train_step,
                                                  self._epoch, 'epoch_started',
                                                  tag=self._current_publish_tag)
            self._epoch_started_marker = True

        # save weights and biases to publish just before current step ends (in step()). this ensures
        # we can publish the proper updates
        if hasattr(module, 'weight'):
            self._weight_cache[module] = module.weight.clone()

        if hasattr(module, 'bias') and module.bias is not None:   # bias can be present, but be None
            self._bias_cache[module] = module.bias.clone()

        self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                             'activations', self._modules[module], out_,
                                             tag=self._current_publish_tag)

    def new_layer_gradients(self, module, gradients):
        '''Callback for newly arriving layer gradients (loss wrt layer output). Registered as a hook
        to the tracked modules.

        .. warning::
            Currently, only layers with one output are supported.

        Parameters
        ----------
        module  :   torch.nn.Module
        gradients    :  torch.Tensor
                        The gradients of the loss w.r.t. layer output

        Raises
        ------
        RuntimeError
            If the module has multiple outputs
        '''
        if isinstance(gradients, tuple):
            if len(gradients) != 1:
                raise RuntimeError('Layers with more than one output are not supported.')
            else:
                gradients = gradients[0]

        self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                             'layer_gradients', self._modules[module], gradients,
                                             tag=self._current_publish_tag)

    def new_parameter_gradients(self, module, gradients):
        '''Callback for newly arriving gradients wrt weight and/or bias. Registered as a hook to the
        tracked modules.  Will trigger export of all new gradient data.

        Parameters
        ----------
        module  :   torch.nn.Module
        gradients    :   tuple(torch.Tensor, torch.Tensor)
                        The gradients w.r.t weight and bias.
        '''
        self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                             'weight_gradients', self._modules[module],
                                             gradients[0],
                                             tag=self._current_publish_tag)

        if gradients[1] is not None:
            self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                                 'bias_gradients', self._modules[module],
                                                 gradients[1],
                                                 tag=self._current_publish_tag)

    def set_model(self, model):
        '''Set the model for direct access for some metrics.

        Parameters
        ----------
        model   :   torch.nn.Module
        '''
        self._model = model
        from types import MethodType
        #############################################
        #  Patch the train function to notify self  #
        #############################################
        train_fn = model.train

        def new_train_fn(this, mode=True):
            train_fn(mode=mode)
            self._is_training = mode
        model.train = MethodType(new_train_fn, model)

        #########################################################
        #  Patch forward function to step() self automatically  #
        #########################################################
        forward_fn = model.forward

        def new_forward_fn(this, *args, should_train=True, tag=self._current_publish_tag):
            '''When subscribers need to push data through the net, they should be given access to
            model.forward() to control the Exporter's behaviour until their compute() method ends.
            The `should_train` parameter can be used to temporarily have the model in validation
            mode. The `tag` parameter can be used to temporarily have all messages be published with
            a different tag.
            '''
            previous_tag = self._current_publish_tag
            self._current_publish_tag = tag
            # In order for accuracy subscribers to not need the model access, we add a secret
            # parameter which they can use to temporarily set the training to False and have it
            # revert automatically. TODO: Check if this is inefficient
            was_training = this.training        # store old value
            this.train(should_train)            # disable/enable training
            if this.training:
                # we need to step before forward pass, else act and grads get different steps
                self.step()
                self.new_input_data(*args)               # do this after stepping
            ret = forward_fn(*args)             # do forward pass w/o messages spawning

            self._current_publish_tag = previous_tag
            this.train(was_training)            # restore previous state
            return ret
        model.forward = MethodType(new_forward_fn, model)

    def set_loss(self, loss_function):
        '''Add hook to loss function to extract labels.

        Parameters
        ----------
        loss_function   :   torch.nn._Loss
        '''

        def hook(mod, output_and_labels, loss):
            network_output, labels = output_and_labels
            self.new_output_and_labels(network_output, labels)
            self.new_loss(loss)

        loss_function.register_forward_hook(hook)

    def step(self):
        '''Increase batch counter (per epoch) and the global step counter.'''
        # due to the fact that backprop happens after forward() was called, we need to step before
        # the forward pass so activation and gradient msgs have the same counter. therefore, the
        # counters start at -1, but we publish a 'batch_finished' message only from the second
        # iteration onwards
        if self._train_step > -1:
            self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                                  'batch_finished',
                                                  tag=self._current_publish_tag)

        self._train_step  += 1
        self._global_step += 1

        for module, weight in self._weight_cache.items():
            self._msg_bus.publish_module_message(self._global_step, self._train_step,
                                                 self._epoch, 'weight_updates',
                                                 self._modules[module],
                                                 module.weight - weight,
                                                 tag=self._current_publish_tag)
            self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                                 'weights', self._modules[module], weight,
                                                 tag=self._current_publish_tag)

        for module, bias in self._bias_cache.items():
            self._msg_bus.publish_module_message(self._global_step, self._train_step,
                                                 self._epoch, 'bias_updates',
                                                 self._modules[module],
                                                 module.bias - bias,
                                                 tag=self._current_publish_tag)
            self._msg_bus.publish_module_message(self._global_step, self._train_step, self._epoch,
                                                 'biases', self._modules[module], bias,
                                                 tag=self._current_publish_tag)

        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'batch_started',
                                              tag=self._current_publish_tag)

    def freeze_module(self, module):
        '''Convenience method for freezing training for a module.

        Parameters
        ----------
        module    :   torch.nn.Module
                      Module to freeze
        '''

        if module not in self._frozen:
            self._frozen.add(module)
            print(f'Freezing {module}')
            freeze_module(module)

    def epoch_finished(self):
        '''Increase the epoch counter and reset the batch counter.'''
        self._msg_bus.publish_network_message(self._global_step, self._train_step, self._epoch,
                                              'epoch_finished',
                                              tag=self._current_publish_tag)
        self._epoch     += 1
        self._train_step = 0
        self._epoch_started_marker = False



'''
.. moduleauthor Rasmus Diederichsen

This module defines :class:`ikkuna.models.VGG`, a small VGG network
'''
import torch.nn as nn


class VGG(nn.Module):
    def __init__(self, input_shape, num_classes=1000, exporter=None):
        super(VGG, self).__init__()
        self.in_channels = input_shape[-1]
        assert 1 <= self.in_channels <= 3, 'Expected 1-3 channels'
        assert input_shape[0] >= 32 and input_shape[1] >= 32, 'Images too small (min 32px per side)'

        self.features = self._make_layers()
        feature_outputs = 512 * (input_shape[0] // 2**5)  * (input_shape[1] // 2**5)
        self.classifier = nn.Linear(feature_outputs, num_classes)

        if exporter:
            exporter.set_model(self)
            exporter.add_modules(self)

    def forward(self, x):
        out = self.features(x)
        out = out.view(out.size(0), -1)
        out = self.classifier(out)
        return out

    def _make_layers(self):
        layers = []
        in_channels = self.in_channels
        for x in [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M']:
            if x == 'M':
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                layers += [nn.Conv2d(in_channels, x, kernel_size=3, padding=1),
                           nn.BatchNorm2d(x),
                           nn.ReLU(inplace=False)]
                in_channels = x
        layers += [nn.AvgPool2d(kernel_size=1, stride=1)]
        return nn.Sequential(*layers)



def seed_everything(seed=1234):
    '''Set the seed for :mod:`torch`, :mod:`random`, :mod:`numpy` and :mod:`torch.cuda`, as well as
    the ``PYTHONHASHSEED`` env var. It also configures CuDNN to use deterministic mode.

    .. warning::

        Setting the CuDNN backend to deterministic mode potentially incurs a performance penalty

    Parameters
    ----------
    seed    :   int
                Seed value to use.
    '''
    import random, torch, os, numpy as np       # noqa
    random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
    np.random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)


def make_fill_polygons(xs, ys, zs):
    '''Make a set of polygons to fill the space below a line in 3d.

    Returns
    -------
    mpl_toolkits.mplot3d.art3d.Poly3DCollection
    '''
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    v = []
    h = 0
    for k in range(0, len(xs) - 1):
        x = [xs[k], xs[k+1], xs[k+1], xs[k]]
        y = [ys[k], ys[k+1], ys[k+1], ys[k]]
        z = [zs[k], zs[k+1],       h,     h]
        v.append(list(zip(x, y, z)))
    poly3dCollection = Poly3DCollection(v)
    return poly3dCollection


def available_optimizers():
    '''List names of all available torch optimizers form :py:mod:`torch.optim`.

    Returns
    -------
    list(str)
        A list of all optimizer names.
    '''
    import torch
    import experiments.optimizers
    # get all module properties which aren't magic
    available_optimizers = set(getattr(torch.optim, name)
                               for name in dir(torch.optim)
                               if not name.startswith('__'))
    available_optimizers = available_optimizers.union(set(getattr(experiments.optimizers, name)
                                                          for name in dir(experiments.optimizers)
                                                          if not name.startswith('__')))
    # remove everything which is a module and not a class (looking at you, lr_scheduler o_o)
    available_optimizers = filter(lambda o: isinstance(o, type), available_optimizers)
    # map them to their class name (w/o module)
    available_optimizers = map(lambda o: o.__name__, available_optimizers)

    return list(available_optimizers)


def create_optimizer(model, name, **kwargs):
    '''Create an optimizer for ``model``\ s parameters. Will disregard all params
    with ``requires_grad == False``.

    Parameters
    ----------
    model   :   torch.nn.Module
    name    :   str
                Name of the optimizer
    kwargs  :   dict
                All arguments which should be passed to the optimizer.

    Raises
    ------
    ValueError
        If superfluous ``kwargs`` are passed.

    '''
    import torch
    import experiments.optimizers
    if name not in available_optimizers():
        raise ValueError(f'Unknown optimizer {name}')

    params = [p for p in model.parameters() if p.requires_grad]
    try:
        return getattr(torch.optim, name)(params, **kwargs)
    except AttributeError:
        return getattr(experiments.optimizers, name)(params, **kwargs)


def get_model(model_name, *args, **kwargs):
    from ikkuna import models
    try:
        if model_name.startswith('ResNet'):
            model_fn = getattr(models, model_name.lower())
            if args:
                print(f'Warning: Ignored args for {model_name} ({args})')
            model = model_fn(**kwargs)
        elif model_name == 'AdamModel':
            from experiments.adam.adam_model import AdamModel
            model = AdamModel(*args, **kwargs)
        else:
            Model = getattr(models, model_name)
            model = Model(*args, **kwargs)
    except AttributeError:
        raise ValueError(f'Unknown model {model}')
    else:
        return model


def initialize_model(module, bias_val=0.01):
    '''Perform weight initialization on `module`. This is somewhat hacky since
    it assumes the presence of `weight` and/or `bias` fields on the module. Will
    skip if not present.

    Parameters
    ----------
    module  :   torch.nn.Module
                The model
    bias_val    :   float
                    Constant for biases (should be small and positive)

    Raises
    ------
    ValueError
        If ``module`` is not one of the known models (currently :class:`~ikkuna.models.AlexNetMini`
        and :class:`~ikkuna.models.DenseNet`)
    '''
    from ikkuna import models
    import torch.nn as nn
    if isinstance(module, models.AlexNetMini):
        for m in module.modules():
            if hasattr(module, 'weight'):
                nn.init.xavier_uniform_(module.weight)
            if hasattr(module, 'bias'):
                module.bias.data.fill_(bias_val)
    elif isinstance(module, models.DenseNet):
        # Official init from torch repo.
        for m in module.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight.data)
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                m.bias.data.zero_()
    else:
        raise ValueError(f'Don\'t know how to initialize {module.__class__.__name__}')


import subprocess


def get_memory_stats(mode='total', unit='mb'):
    '''Get the total|used|free gpu memory.

    Parameters
    ----------
    mode    :   str
                One of ``'total'``, ``'used'`` or ``'free'``

    Returns
    -------
    dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    '''
    if mode not in ['total', 'used', 'free']:
        raise ValueError

    units = ['b', 'kb', 'mb', 'gb']
    factors = [1024 ** i for i in range(len(units))]
    if unit not in units:
        raise ValueError(f'Unknown unit "{unit}"')

    memory = subprocess.check_output(
        [
            'nvidia-smi', f'--query-gpu=memory.{mode}',
            '--format=csv,nounits,noheader'
        ], encoding='utf-8')
    # Convert lines into a dictionary
    index = units.index(unit)
    # nvidia-smi gives mb, so go back to b and then up to the selected unit
    size = [int(x) * factors[2] / factors[index] for x in memory.strip().split('\n')]
    return dict(zip(range(len(size)), size))


import torchvision
from torchvision.transforms import ToTensor, Compose
import numpy as np
import torch


def _load_imagenet_dogs(root, formats, train_transforms, test_transforms):
    '''Load the ImageNetDogs dataset from disk'''
    from torchvision.datasets import DatasetFolder
    from torch.utils.data import random_split
    from PIL import Image
    import copy

    def loader(path):
        '''Load image from fs path'''
        return Image.open(path)

    dataset = DatasetFolder(root, loader, formats)

    size                        = len(dataset)
    size_train                  = int(0.7 * size)
    size_test                   = size - size_train
    dataset_train, dataset_test = random_split(dataset, [size_train, size_test])

    # set the transforms
    # we need to make a copy for the test set to have different transforms
    dataset_test.dataset            = copy.copy(dataset)
    dataset_train.dataset.transform = train_transforms
    dataset_test.dataset.transform  = test_transforms
    return dataset_train, dataset_test


def load_dataset(name, train_transforms=None, test_transforms=None, **kwargs):
    '''Retrieve a dataset and determine the number of classes. This estimate is
    obtained from the number of different values in the training labels.

    Parameters
    ----------
    name    :   str
                Currently, dataset names in :mod:`torchvision.datasets` and ``ImageNetDogs`` are
                supported.
    train_transforms    :   list
                            List of transforms on the train data. Defaults to
                            :class:`torchvision.transforms.ToTensor`
    test_transforms    :   list
                            List of transforms on the test data. Defaults to
                            :class:`torchvision.transforms.ToTensor`

    Keyword Arguments
    -----------------
    root    :   str
                Root directory for dataset folders. Defaults to
                ``/home/share/software/data/<name>/Images``
    formats :   list
                List of file extensions for dataset folders. Defaults to ``['jpg', 'png']``

    Returns
    -------
    tuple
        2 :class:`~train.DatasetMeta`\ s are returned, one for train and one test set
    '''
    def identity(img):
        return img

    train_transforms = Compose(train_transforms) if train_transforms else identity
    test_transforms  = Compose(test_transforms) if test_transforms else identity

    ##########################################
    #  Get the datasets in train/test split  #
    ##########################################
    if name == 'ImageNetDogs':
        root    = kwargs.get('root', f'/home/share/software/data/{name}/Images/')
        formats = kwargs.get('formats', ['jpg', 'png'])
        dataset_train, dataset_test = _load_imagenet_dogs(root, formats, train_transforms,
                                                          test_transforms)
    else:
        try:
            if name == 'WhitenedCIFAR10':
                from experiments.adam.cifar_dataset import WhitenedCIFAR10
                dataset_cls = WhitenedCIFAR10
            else:
                dataset_cls   = getattr(torchvision.datasets, name)

            import os
            if os.path.exists('/home/share/'):
                path = '/home/share/data'
            else:
                path = '/tmp/data'
            dataset_train = dataset_cls(path,
                                        download=True,
                                        train=True,
                                        transform=train_transforms
                                        )
            dataset_test  = dataset_cls(path,
                                        download=True,
                                        train=False,
                                        transform=test_transforms
                                        )
        except AttributeError:
            raise NameError(f'Dataset {name} unknown.')

    ########################
    #  Determine metadata  #
    ########################
    def num_classes(dataset):
        if hasattr(dataset, 'dataset'):     # is a Subset
            dataset = dataset.dataset
        if hasattr(dataset, 'targets'):
            labels = dataset.targets
        elif hasattr(dataset, 'labels'):
            labels = dataset.labels
        else:
            raise RuntimeError(f'{dataset} has neither `targets` nor `labels` properties.')


        # infer number of classes from labels. will fail if not all classes occur in labels
        if isinstance(labels, np.ndarray):
            return np.unique(labels).size()
        elif isinstance(labels, list):
            return len(set(labels))
        elif isinstance(labels, torch.Tensor):
            return labels.unique().numel()
        else:
            raise ValueError(f'Unexpected label storage {labels.__class__.__name__}')

    def shape(dataset):
        N = len(dataset)
        sample, _ = next(iter(dataset))

        # if only three dimensions, assume [N, H, W], else [N, H, W, C]
        if sample.ndimension() == 3:
            C, H, W = sample.shape
        else:
            H, W = sample.shape
            C = 1
        return (N, H, W, C)

    from train import DatasetMeta
    meta_train = DatasetMeta(dataset=dataset_train, num_classes=num_classes(dataset_train),
                             shape=shape(dataset_train))
    meta_test  = DatasetMeta(dataset=dataset_test, num_classes=num_classes(dataset_test),
                             shape=shape(dataset_test))

    return meta_train, meta_test


def freeze_module(module):
    def freeze(m):
        for p in m.parameters():
            m.requires_grad = False

    module.apply(freeze)



from .alexnetmini import AlexNetMini
from .densenet import DenseNet
from .resnet import ResNet
from .resnet import resnet18, resnet34, resnet50, resnet101, resnet152
from .vgg import VGG
from .fc import FullyConnectedModel

__all__ = ['AlexNetMini', 'DenseNet', 'ResNet', 'VGG', 'FullyConnectedModel']


def get_model(name, *args, **kwargs):
    '''Obtain a model instance by name.

    Parameters
    ----------
    name    :   str
                One of the class names of the models defined in this package, or ``resnet<dd>`` for
                the different ResNets

    Other arguments
    ---------------
    Further positional args and kwargs are passed to the model initialiser, except for ResNets,
    which only receive the keyword args.
    '''
    try:
        if name.startswith('ResNet'):
            model_fn = globals()[name.lower()]
            model    = model_fn(**kwargs)
        else:
            Model    = globals()[name]
            model    = Model(*args, **kwargs)
        return model
    except AttributeError:
        raise ValueError(f'Unknown model {model}')



import torch

from ikkuna.export.subscriber import PlotSubscriber, SynchronizedSubscription
from ikkuna.export.messages import get_default_bus

ZERO_TENSOR = torch.tensor(0.0, device='cuda' if torch.cuda.is_available() else 'cpu')


class RatioSubscriber(PlotSubscriber):
    '''A :class:`~ikkuna.export.subscriber.Subscriber` which computes the average ratio between two
    quantities.  The dividend will be the first element of the
    :attr:`~ikkuna.export.subscriber.Subscription.kinds` property, the divisor the second.
    Therefore it is vital to pass the message kinds to the
    :class:`~ikkuna.export.subscriber.Subscription` object in the correct order.'''

    def __init__(self, kinds, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb', absolute=True):
        '''
        Parameters
        ----------
        absolute :  bool
                    Whether to use absolute ratio
        '''
        if len(kinds) != 2:
            raise ValueError(f'RatioSubscriber requires 2 kinds, got {len(kinds)}.')

        title        = f'{kinds[0]}_{kinds[1]}_ratio'
        ylabel       = 'Ratio'
        xlabel       = 'Train step'
        subscription = SynchronizedSubscription(self, kinds, tag=tag, subsample=subsample)
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)
        if absolute:
            self._metric_postprocess = torch.abs
        else:
            self._metric_postprocess = lambda x: x

        self._add_publication(f'{kinds[0]}_{kinds[1]}_ratio', type='DATA')

    def compute(self, message_bundle):
        '''The ratio between the two kinds is computed as the ratio of L2-Norms of the two Tensors.
        A :class:`~ikkuna.export.messages.ModuleMessage` with the identifier
        ``{kind1}_{kind2}_ratio`` is published.'''

        module, module_name = message_bundle.key

        dividend            = message_bundle.data[message_bundle.kinds[0]]
        divisor             = message_bundle.data[message_bundle.kinds[1]]

        scale1              = dividend.norm()
        scale2              = divisor.norm()
        ratio               = (scale1 / scale2).item()

        self._backend.add_data(module_name, ratio, message_bundle.global_step)

        kind = f'{message_bundle.kinds[0]}_{message_bundle.kinds[1]}_ratio'
        self.message_bus.publish_module_message(message_bundle.global_step,
                                                message_bundle.train_step,
                                                message_bundle.epoch, kind,
                                                message_bundle.key, ratio)



from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class VarianceSubscriber(PlotSubscriber):
    '''A :class:`~ikkuna.export.subscriber.Subscriber` which computes the
    variance of quantity for the current batch.
    '''

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):

        if not isinstance(kind, str):
            raise ValueError('VarianceSubscriber only accepts 1 kind')

        title        = f'{kind}_variance'
        ylabel       = 'σ^2'
        xlabel       = 'Train step'
        subscription = Subscription(self, [kind], tag=tag, subsample=subsample)
        super().__init__([subscription], message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)

        self._add_publication(f'{kind}_variance', type='DATA')

    def compute(self, message):
        '''Compute the variance of a quantity. A :class:`~ikkuna.export.messages.ModuleMessage`
        with the identifier ``{kind}_variance`` is published. '''

        module, module_name = message.key
        data                = message.data
        var                 = data.var()
        self._backend.add_data(module_name, var, message.global_step)

        kind = f'{message.kind}_variance'
        self.message_bus.publish_module_message(message.global_step,
                                                message.train_step,
                                                message.epoch, kind,
                                                message.key, var)



from collections import defaultdict
import numpy as np
import torch
from torch.utils.data import Subset, DataLoader
import svcca

from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus
from ikkuna.utils import freeze_module


class ChunkedDict(object):
    '''A dictionary-like class that can be used to incrementally fill tensors of predetermined size
    by accumulating data. The dictionary keeps track of what has already been received and kan thus
    be asked if all data for some module was received.  This can be used e.g. for stitching together
    an activation matrix of (n_neurons, n_datapoints) when the activations arrive in batches.'''

    def __init__(self, expected_n):
        '''
        Parameters
        ----------
        expected_n  :   int
                        The number of datapoints to expect for all modules
        '''
        self._expected_n = expected_n
        self._received_n = defaultdict(int)
        self._data = dict()

    def clear(self):
        '''Remove all entries.'''
        self._received_n.clear()
        self._data.clear()

    def append(self, module, data):
        '''Add data for a module. Creates new record if one is not already open for this module.

        Parameters
        ----------
        module  :   torch.nn.Module or str
                    Some unique key for this module
        data    :   torch.Tensor
                    Data tensor to append
        '''
        if module not in self._data:
            self._data[module] = torch.zeros(self._expected_n, *data.shape[1:], device=data.device)

        n = data.shape[0]
        received = self._received_n[module]
        self._data[module][received:received + n, ...] = data
        self._received_n[module] += n

    def keys(self):
        '''Get all known modules, completed or otherwise.

        Returns
        -------
        dict_keys
            View of the data dictionary's keys
        '''
        return self._data.keys()

    def bytesize(self):
        '''Compute occupied bytes of the data (without the keys)

        Returns
        -------
        int
            Bytes consumed by all the tensors
        '''
        size = 0
        for module in self._data:
            size += self._data[module].numel() * 4
        return size

    def __setitem__(self, module, data):
        '''Set the data for a module, overwriting any previous data.'''
        if module in self._data:
            del self._data[module]
            del self._received_n[module]
        self.append(module, data)

    def __getitem__(self, key):
        '''Get data for a module.

        Returns
        -------
        torch.Tensor

        Raises
        ------
        KeyError
            If the module's data is not yet complete
        '''
        if not self.complete(key):
            raise KeyError(f'Data for {key} not yet complete.')
        return self._data[key]

    def __contains__(self, module):
        return module in self._data

    def complete(self, module):
        '''Check if expected number of datapoints have been received for a module.

        Returns
        -------
        bool
        '''
        return self._received_n[module] == self._expected_n

    def pop(self, module):
        '''Remove and return data for module'''
        data = self[module]
        del self._data[module]
        del self._received_n[module]
        return data


class BatchedSVCCASubscriber(PlotSubscriber):
    '''A subscriber which at some interval halts training, propagates some dataset through the net,
    records activations for all modules and stores them. At the next checkpoint, SVCCA similarity
    is computed between every layer at time step t and the same layer at t-1.

    .. note::
        It seems that despite the fact that SVCCA runs at least twice as fast on the GPU, this
        subscriber is not faster than on the CPU. I'm not sure why that is, since I'm even caching
        input data to avoid repeatedly copying to the GPU

    .. warning::
        Currently, this subscriber uses Numpy for SVCAA. For some reason, the activations produced
        by networks lead to highly ill-conditioned covariance matrices and PyTorch's SVD handles
        them differently than Numpy does. SVD does not converge in that case, and
        ``robust_cca_similarity`` must be used, which defeats the whole idea of speeding up the
        process.
    '''

    def __init__(self, dataset_meta, n, forward_fn, freeze_at=10, batch_size=256,
                 message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):
        '''
        Parameters
        ----------
        dataset_meta    :   train.DatasetMeta
                            Dataset to load data from for retrieving activations
        n   :   int
                Number of datapoints to randomly sample
        freeze_at   :   float
                        Similarity threshold after which a layer is frozen. Values >= 1 will turn
                        off freezing
        batch_size  :   int
                        Batch size to use for forward passes. Can be set to some value if the entire
                        data at once would be too larger. Otherwise use :class:`SVCCASubscriber`
        '''

        self._forward_fn     = forward_fn
        self._previous_acts  = ChunkedDict(n)
        self._current_acts   = ChunkedDict(n)
        indices              = np.random.randint(0, dataset_meta.size, size=n)
        dataset              = Subset(dataset_meta.dataset, indices)
        self._loader         = DataLoader(dataset, batch_size=batch_size, shuffle=False,
                                          pin_memory=True)
        # cache tensors so we don't repeatedly deserialize and copy
        self._input_cache    = []

        self._freeze_at      = freeze_at if isinstance(freeze_at, float) else 10.
        self._ignore_modules = set()

        title        = f'self_similarity'
        ylabel       = 'Similarity'
        xlabel       = 'Train step'
        subscription1 = Subscription(self, ['batch_finished'], tag=tag, subsample=subsample)
        subscription2 = Subscription(self, ['activations'], tag='svcca_testing',
                                     subsample=1)
        super().__init__([subscription1, subscription2],
                         message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)
        self._add_publication(f'self_similarity', type='DATA')

    def _module_complete_previous(self, module):
        '''Check if activations for module are completely buffered from the previous step.'''
        return not (module not in self._previous_acts
                    or not self._previous_acts.complete(module))

    def _module_complete_current(self, module):
        '''Check if activations for module are completely buffered from the current step.'''
        return not (module not in self._current_acts
                    or not self._current_acts.complete(module))

    def _record_activations_previous(self, module, data):
        '''Record activations into the ``previous`` buffer'''
        self._previous_acts.append(module, data)

    def _record_activations_current(self, module, data):
        '''Record activations into the ``current`` buffer'''
        self._current_acts.append(module, data)

    def _do_forward_pass(self):
        # cache inputs initially
        if not self._input_cache:
            loader = iter(self._loader)
        else:
            loader = iter(self._input_cache)

        for i, (X, labels) in enumerate(loader):
            X = X.cuda()
            if len(self._input_cache) < i + 1:
                self._input_cache.append((X, labels))
            self._forward_fn(X, should_train=False, tag='svcca_testing')

    def _compute_similarity(self, name):
        # now both current and previous acts are complete and we can compute
        # print(f'Computing similarity for {name}')
        previous_acts = self._previous_acts.pop(name)
        current_acts = self._current_acts.pop(name)

        self._previous_acts[name] = current_acts

        if previous_acts.ndimension() > 2:
            c = previous_acts.shape[1]  # channel dim
            previous_acts = previous_acts.permute([0, 2, 3, 1]).reshape(-1, c)

        if current_acts.ndimension() > 2:
            c = current_acts.shape[1]  # channel dim
            current_acts = current_acts.permute([0, 2, 3, 1]).reshape(-1, c)

        previous_acts = previous_acts.detach().cpu().numpy()
        current_acts = current_acts.detach().cpu().numpy()
        result_dict = svcca.cca_core.robust_cca_similarity(previous_acts.T,
                                                           current_acts.T,
                                                           epsilon=1e-8,
                                                           threshold=0.98,
                                                           verbose=False,
                                                           compute_dirns=False,
                                                           rescale=True)
        return result_dict['mean'][0]

    def _freeze_module(self, module):
        print(f'Freezing {module}')
        freeze_module(module)
        self._ignore_modules.add(module)
        if module in self._previous_acts:
            self._previous_acts.pop(module)
        if module in self._current_acts:
            self._current_acts.pop(module)

    def compute(self, message):
        '''A :class:`~ikkuna.export.messages.NetworkMessage` with the identifier ``self_similarity``
        will be published.'''

        if message.tag == 'default' and message.kind == 'batch_finished':
            self._do_forward_pass()

        elif message.tag == 'svcca_testing':
            module, name = message.key
            if module in self._ignore_modules:
                return

            if not self._module_complete_previous(module):
                self._record_activations_previous(module, message.data)
            elif not self._module_complete_current(module):
                self._record_activations_current(module, message.data)

            if self._module_complete_current(module) and self._module_complete_previous(module):
                mean = self._compute_similarity(module)
                self._backend.add_data(name, mean, message.global_step)
                self.message_bus.publish_module_message(message.global_step,
                                                        message.train_step,
                                                        message.epoch,
                                                        'self_similarity',
                                                        message.key,
                                                        data=mean)

                if mean > self._freeze_at:
                    self._freeze_module(module)


class SVCCASubscriber(BatchedSVCCASubscriber):
    '''Simplified :class:`BatchedSVCCASubscriber` subclass which does not use a
    ``ChunkedDict`` but instead regular dicts and thus assumes that the entire test data can be
    propagated through the model in one go and batching is unnecessary.'''

    def __init__(self, dataset_meta, n,  *args, **kwargs):
        super().__init__(dataset_meta, n, *args, **kwargs, batch_size=n)
        self._previous_acts = dict()
        self._current_acts = dict()

    def _module_complete_previous(self, module):
        return module in self._previous_acts

    def _module_complete_current(self, module):
        return module in self._current_acts

    def _record_activations_previous(self, module, data):
        self._previous_acts[module] = data

    def _record_activations_current(self, module, data):
        self._current_acts[module] = data



import torch
from torch.nn.functional import normalize

from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class ConditionNumberSubscriber(PlotSubscriber):
    '''A Subscriber which computes the condition number of a matrix. Not sure what this is useful
    for, but maybe for Linear layers.'''

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):
        '''
        Parameters
        ----------
        kind    :   str
                    Message kind to compute condition number norm on. Not sure if it makes sense
                    for non-2d matrices, which have to be reshaped to 2-d
                    non-matrix type.


        For other parameters, see :class:`~ikkuna.export.subscriber.PlotSubscriber`
        '''

        if not isinstance(kind, str):
            raise ValueError(f'{self.__class__.__name__} only accepts 1 kind')

        subscription = Subscription(self, [kind], tag=tag, subsample=subsample)

        title = f'{kind}_condition_number'
        xlabel = 'Step'
        ylabel = 'Condition number'
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'xlabel': xlabel,
                          'ylims': ylims,
                          'ylabel': ylabel},
                         backend=backend)
        self._add_publication(f'{kind}_condition_number', type='DATA')
        self.u = dict()
        self.u_inv = dict()

    def compute(self, message):
        '''A :class:`~ikkuna.export.messages.ModuleMessage`
        with the identifier ``{kind}_condition_number`` is published. '''

        module, module_name = message.key
        # get and reshape the weight tensor to 2d
        weights             = message.data
        height              = weights.size(0)
        weights2d           = weights.reshape(height, -1)
        width               = weights2d.size(1)
        weights2d_inv       = torch.pinverse(weights2d)

        # buffer for power iteration (don't know what the mahematical purpose is)
        if module_name not in self.u:
            self.u[module_name] = normalize(weights2d.new_empty(height).normal_(0, 1), dim=0)
            self.u_inv[module_name] = normalize(weights2d_inv.new_empty(width).normal_(0, 1), dim=0)

        # estimate singular values
        with torch.no_grad():
            for _ in range(3):      # TODO: Make niter parameter
                v = normalize(torch.matmul(weights2d.t(), self.u[module_name]), dim=0)
                self.u[module_name] = normalize(torch.matmul(weights2d, v), dim=0)

                v_inv = normalize(torch.matmul(weights2d_inv.t(), self.u_inv[module_name]), dim=0)
                self.u_inv[module_name] = normalize(torch.matmul(weights2d_inv, v_inv), dim=0)

        norm = torch.dot(self.u[module_name], torch.matmul(weights2d, v)).item()
        norm_inv = torch.dot(self.u_inv[module_name], torch.matmul(weights2d_inv, v_inv)).item()

        condition = norm / norm_inv

        self._backend.add_data(module_name, condition, message.global_step)

        kind = f'{message.kind}_condition_number'
        self.message_bus.publish_module_message(message.global_step,
                                                message.train_step,
                                                message.epoch, kind,
                                                message.key, condition)



from tensorboardX import SummaryWriter
import abc
import matplotlib.pyplot as plt
import numpy as np
from collections import defaultdict
from mpl_toolkits.mplot3d import Axes3D     # noqa
from ikkuna.utils import make_fill_polygons
import torch


class Backend(abc.ABC):
    '''Base class for visualiation backends. :class:`~ikkuna.export.subscriber.Subscriber`\ s use
    this class to dispatch their metrics to have them visualised.

    Attributes
    ----------
    title   :   str
                The figure title
    '''

    def __init__(self, title):
        '''
        Parameters
        ----------
        title   :   str
                    Title to use for the figure.
        '''
        self._title = title

    @abc.abstractmethod
    def add_data(self, module_name, datum, step):
        '''Display scalar data (i.e. a line plot)

        Parameters
        ----------
        module_name  :  str
                        Name of module which emitted the data
        datum   :   torch.Tensor
                    Payload
        step    :   int
                    Global step
        '''
        pass

    @abc.abstractmethod
    def add_histogram(self, module_name, datum, step):
        '''Display histogram data (i.e. a line plot)

        Parameters
        ----------
        module_name  :  str
                        Name of module which emitted the data
        datum   :   torch.Tensor
                    Payload, not the histogram itself
        step    :   int
                    Global step
        '''
        pass

    @property
    def title(self):
        return self._title


class NullBackend(Backend):
    def __init__(self, *args, **kwargs):
        pass

    def add_data(self, *args, **kwargs):
        pass

    def add_histogram(self, *args, **kwargs):
        pass

    def title(self):
        return ''



class UpdatableHistogram(object):
    '''A utility class to wrap a single histogram subplot of a figure which can display up to a
    number of histograms. New data displaces old one. This emulates tensorboard histogram traces
    with matplotlib.

    Attributes
    ----------
    _max_hists  :   int
                    Max number of histograms to display before discarding old ones
    _unit   :   float
                Base length used for spacing successive histograms along the y axis
    _fig    :   matplotlib.figure.Figure
                Figure to put subplot in
    _ax :   mpl_toolkits.mplot3d.axes3d.Axes3D
            3d axes object to plot in
    _base_r :   float
                start red value
    _base_g :   float
                start green value
    _base_b :   float
                start blue value
    _hists  :   list
                Stack of currently displayed histograms
    _title  :   str
                Plot title
    '''

    def __init__(self, figure, subplot_conf=111, title='', max_hists=10,
                 basecolor=(199/255, 64/255, 24/255)):
        '''
        Parameters
        ----------
        figure  :   matplotlib.figure.Figure
                    Figure to add subplot to
        subplot_conf    :   int or tuple
                            Subplot configuration. See :func:`matplotlib.pyplot.subplots`
        title   :   str
                    Subplot title
        max_hists   :   int
                        Number of histograms to keep
        basecolor   :   tuple
                        RGB tuple denoting the color of the most recent histogram.
        '''
        if isinstance(subplot_conf, int):
            subplot_conf = (subplot_conf,)
        self._max_hists = max_hists
        self._unit      = 1 / max_hists
        self._fig       = figure
        self._ax        = self._fig.add_subplot(*subplot_conf, projection='3d')
        self._hists     = []
        self._title     = title
        self._base_r, self._base_g, self._base_b = basecolor
        self._prepare_plot()

    def _prepare_plot(self):
        '''Prepare the axes object for (re)plotting. This removes all unnecessary elements and
        assigns labels.'''

        transparent = (1, 1, 1, 0)
        self._ax.set_title(self._title)
        self._ax.set_xlabel('')
        self._ax.set_zlabel('frequency')
        self._ax.get_yaxis().line.set_color(transparent)
        self._ax.set_yticks([])
        self._ax.grid(False)
        self._ax.w_xaxis.set_pane_color(transparent)
        self._ax.w_yaxis.set_pane_color(transparent)
        self._ax.w_zaxis.set_pane_color(transparent)

    @property
    def figure(self):
        return self._fig

    @property
    def ax(self):
        return self._ax

    def replot(self):
        '''Recompute all visible plots. This will remove all current plots and simply replot
        ``self._hists``.

        .. note::

            This is not the most efficient way to update the plot maybe; the proper way to do this
            could involve moving all current histograms one step backwards, unless the x axis
            changes.
        '''
        self._ax.clear()
        self._prepare_plot()
        for i, (hist, edges) in enumerate(reversed(self._hists)):   # last is most recent
            xpos   = edges[:-1]     # take start edge of every bin for x position
            # TODO: For uniform bins, use mean of bin for xpos
            ypos   = i * self._unit * np.ones_like(edges[:-1])      # move according to recency
            height = hist

            # color gets faded for older hists
            percent = i / len(self._hists)
            color = (self._base_r + percent * (1.0 - self._base_r),
                     self._base_g + percent * (1.0 - self._base_g),
                     self._base_b + percent * (1.0 - self._base_b))
            self._ax.plot(xpos, ypos, height, color='black', linewidth=0)

            # fill below the plot with polygons
            collection = make_fill_polygons(xpos, ypos, height)
            collection.set_facecolor(color)
            collection.set_edgecolor(color)
            self._ax.add_collection3d(collection)

    def add_data(self, X):
        '''Add data for a new histogram. Currently, the number of bins is fixed at 50. Old data is
        deleted.

        Parameters
        ----------
        X   :   list
                Arbitrary sequence of tensors to merge for a histogram
        '''
        import torch
        X           = torch.cat(X).detach().cpu().numpy()
        hist, edges = np.histogram(X, bins=50, density=True)
        self._hists.append((hist, edges))
        if len(self._hists) > self._max_hists:
            self._hists.pop(0)


class MPLBackend(Backend):
    '''Matplotlib backend (use in Jupyter with ``%matplotlib inline`` or via X-forwarding over ssh
    [barely useable])

    Attributes
    ----------
    _xlabel :   str
                X axis label for line plots
    _ylabel :   str
                Y axis label for line plots
    _ylims  :   tuple
                Limits of the y axis for line plots
    _redraw_counter :   int
                        Number of datapoints to consume before redrawing the figure
    _plots  :   dict
                module-plot mapping
    _axes   :   dict
                module-UpdatableHistogram mapping (this should be refactored)
    _buffer :   dict
                Per-module buffer of tensors for more reliable histograms
    _buffer_lim :   int
                    Size of the buffer
    '''

    def __init__(self, **kwargs):
        '''
        Parameters
        ----------
        xlabel  :   str
        ylabel  :   str
        ylims   :   tuple
        buffer_lim  :   int
                        Buffer size for more reliable histograms
        '''
        super().__init__(kwargs.get('title'))

        ############################
        #  Line plot/general args  #
        ############################
        self._xlabel          = kwargs.get('xlabel')
        self._ylabel          = kwargs.get('ylabel')
        self._ylims           = kwargs.get('ylims')
        self._redraw_interval = kwargs.get('redraw_interval', 10)
        self._redraw_counter  = 0
        self._figure          = self._ax = None
        self._plots           = None
        self._axes            = None

        ####################
        #  Histogram args  #
        ####################
        self._buffer      = defaultdict(list)
        self._buffer_lim  = kwargs.get('buffer_size', 20)

    def _prepare_axis(self, ax):
        '''Prepare the line plot axis with labels and scaling.'''
        ax.set_xlabel(self.xlabel)
        ax.set_ylabel(self.ylabel)
        ax.set_autoscaley_on(True)

    @Backend.title.setter
    def title(self, title):
        assert len(title) > 0
        self._title = title

    @property
    def xlabel(self):
        return self._xlabel

    @xlabel.setter
    def xlabel(self, label):
        self._xlabel = label

    @property
    def ylabel(self):
        return self._ylabel

    @ylabel.setter
    def ylabel(self, label):
        self._ylabel = label

    def _reflow_plots(self):
        '''Reqorganize the histogram subplots into a rectangular shape. For now, the sublots are
        arranged on a grid twice as high as it is wide, since vertical space is often in
        abundance.

        .. note::

            This does not recompute the plots
        '''
        nplots = len(self._axes)
        h = int(np.floor(np.sqrt(nplots) * 2 + 0.5))    # remove the factor 2 for square grid
        w = int(np.ceil(np.sqrt(nplots) / 2))
        assert h * w >= nplots
        for i, (module, axis) in enumerate(self._axes.items()):
            if isinstance(axis, UpdatableHistogram):
                axis = axis.ax      # TODO: make less ugly
            axis.change_geometry(h, w, i + 1)

    def add_histogram(self, module_name, datum, step):

        if not self._figure:
            # first time? initialise
            self._figure = plt.figure(figsize=(8, 20))
            self._figure.suptitle(self.title)
            self._axes = {}

        if module_name not in self._axes:
            # haven't seen this module_name before? make new plot for it
            nplots = len(self._axes)
            self._axes[module_name] = UpdatableHistogram(self._figure,
                                                         subplot_conf=(nplots + 1, 1, 1),
                                                         title=module_name)

            self._reflow_plots()

        self._buffer[module_name].append(datum)

        if len(self._buffer[module_name]) == self._buffer_lim:   # buffer full
            self._axes[module_name].add_data(self._buffer[module_name])
            self._axes[module_name].replot()
            self._buffer[module_name] = []

            self._figure.canvas.draw()
            self._figure.canvas.flush_events()
            self._figure.show()

    def add_data(self, module_name, datum, step):

        if not self._figure:
            self._figure, self._ax = plt.subplots()
            self._figure.suptitle(self.title)
            self._plots            = {}
            self._prepare_axis(self._ax)

        if module_name not in self._plots:
            # empty plot so we can simply set the data later
            self._plots[module_name] = self._ax.plot([], [], label=f'{module_name}')[0]

        xdata = self._plots[module_name].get_xdata()
        ydata = self._plots[module_name].get_ydata()
        xdata = np.append(xdata, 1 if len(xdata) == 0 else xdata[-1] + 1)
        if isinstance(datum, torch.Tensor):
            datum = datum.detach().cpu().numpy()
        ydata = np.append(ydata, datum)

        self._plots[module_name].set_xdata(xdata)
        self._plots[module_name].set_ydata(ydata)
        self._ax.legend(ncol=2)

        if self._redraw_counter % self._redraw_interval == 0:
            self._ax.relim()
            if self._ylims:
                self._ax.set_ylim(self._ylims)
            self._ax.autoscale_view(scaley=True)

            # redraw the figure
            self._figure.canvas.draw()
            self._figure.canvas.flush_events()
            self._figure.show()
            self._redraw_counter = 0

        self._redraw_counter += 1


import functools


# use lru cache to generate new result only once per experiment run
# TODO: remove this hack
@functools.lru_cache(None)
def determine_run_index(log_dir):
    import os
    if not os.path.exists(log_dir):
        return 0
    else:
        subdirs = os.listdir(log_dir)
        new_index = len(subdirs)
        return new_index


prefix = None


def configure_prefix(p):
    '''Set a prefix to the log directory for Tensorboard.

    Parameters
    ----------
    p   :   str
            Prefix to the directory name. ``_runs`` will be appended by :class:`TBBackend`
    '''
    global prefix
    prefix = p


def set_run_info(info):
    TBBackend.info = info


class TBBackend(Backend):
    '''Tensorboard backend.

    .. note::

        Whitespace and punctuation in the ``title`` will be replaced
        with underscores due to the fact that it becomes part of a file name.

    Attributes
    ----------
    _writer :   tensorboardX.SummaryWriter
    _hist_bins   :   int
                     Number of bins to use for histograms
    '''

    # TODO: make printing metadata non-hacky
    info : str = 'n/a'

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop('title'))
        self._hist_bins = kwargs.pop('bins', 50)
        log_dir         = kwargs.pop('log_dir', 'runs' if not prefix else prefix)
        index           = determine_run_index(log_dir)
        log_dir         = f'{log_dir}/run{index}'
        self._writer    = SummaryWriter(log_dir)
        self._writer.add_text('run_conf', TBBackend.info)

    def add_data(self, module_name, datum, step):
        # Unfortunately, xlabels, ylabels and plot titles are not supported
        self._writer.add_scalars(f'{self.title}', {module_name: datum}, global_step=step)

    def add_histogram(self, module_name, datum, step, bins='auto'):
        self._writer.add_histogram(f'{self.title}: {module_name}', datum, global_step=step,
                                   bins=bins)



import importlib
import pkg_resources

from .subscriber import (Subscriber, Subscription, SynchronizedSubscription, PlotSubscriber,
                         CallbackSubscriber)

__all__ = ['Subscriber', 'Subscription', 'SynchronizedSubscription', 'PlotSubscriber',
           'CallbackSubscriber']

####################################################################################################
#                                             PLUGINS                                              #
####################################################################################################

# Discover all installed plugins at the entry point
subscribers = {
    entry_point.name: entry_point.load()
    for entry_point in pkg_resources.iter_entry_points('ikkuna.export.subscriber')
}
__all__.extend(list(subscribers.keys()))


# use importlib to add them to the module's properties, so that we can write
# `ikkuna.export.subscriber.CustomSubscriber`, even if the class was defined in
# `some_module.foobar.custom.py`. How else would plugins make sense?
# This seems like a hack.
for name, clazz in subscribers.items():
    locals()[clazz.__name__] = getattr(importlib.import_module(clazz.__module__), clazz.__name__)



import torch
from torch.nn.functional import normalize

from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class SpectralNormSubscriber(PlotSubscriber):

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):
        '''
        Parameters
        ----------
        kind    :   str
                    Message kind to compute spectral norm on. Doesn't make sense with kinds of
                    non-matrix type.


        For other parameters, see :class:`~ikkuna.export.subscriber.PlotSubscriber`
        '''

        if not isinstance(kind, str):
            raise ValueError('SpectralNormSubscriber only accepts 1 kind')

        subscription = Subscription(self, [kind], tag, subsample)

        title = f'{kind}_spectral_norm'
        xlabel = 'Step'
        ylabel = 'Spectral norm'
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'xlabel': xlabel,
                          'ylims': ylims,
                          'ylabel': ylabel},
                         backend=backend)
        self.u = dict()
        self._add_publication(f'{kind}_spectral_norm', type='DATA')

    def compute(self, message):
        '''The spectral norm computation is taken from the `Pytorch implementation of spectral norm
        <https://pytorch.org/docs/master/_modules/torch/nn/utils/spectral_norm.html>`_. It's
        possible to use SVD instead, but we are not interested in the full matrix decomposition,
        merely in the singular values.

        A :class:`~ikkuna.export.messages.ModuleMessage`
        with the identifier ``{kind}_spectral_norm`` is published. '''

        module, module_name = message.key
        # get and reshape the weight tensor to 2d
        weights             = message.data
        height              = weights.size(0)
        weights2d           = weights.reshape(height, -1)

        # buffer for power iteration (don't know what the mahematical purpose is)
        if module_name not in self.u:
            self.u[module_name] = normalize(weights2d.new_empty(height).normal_(0, 1), dim=0)

        # estimate singular values
        with torch.no_grad():
            for _ in range(3):      # TODO: Make niter parameter
                v = normalize(torch.matmul(weights2d.t(), self.u[module_name]), dim=0)
                self.u[module_name] = normalize(torch.matmul(weights2d, v), dim=0)

        norm = torch.dot(self.u[module_name], torch.matmul(weights2d, v)).item()

        self._backend.add_data(module_name, norm, message.global_step)

        kind = f'{message.kind}_spectral_norm'
        self.message_bus.publish_module_message(message.global_step,
                                                message.train_step,
                                                message.epoch, kind,
                                                message.key, norm)



from ikkuna.export.subscriber import PlotSubscriber, SynchronizedSubscription
from ikkuna.export.messages import get_default_bus


class TrainAccuracySubscriber(PlotSubscriber):
    ''':class:`~ikkuna.export.subscriber.Subscriber` which computes the batch accuracy.'''

    def __init__(self, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):
        '''
        For parameters see :class:`~ikkuna.export.subscriber.PlotSubscriber`
        '''
        subscription = SynchronizedSubscription(self, ['network_output', 'input_labels'], tag=tag,
                                                subsample=subsample)

        title  = f'train_batch_accuracy'
        xlabel = 'Step'
        ylabel = 'Accuracy'
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'xlabel': xlabel,
                          'ylims': ylims,
                          'ylabel': ylabel},
                         backend=backend)

        self._add_publication(f'train_accuracy', type='META')

    def compute(self, message):
        '''Compute accuracy over the current train batch.

        A :class:`~ikkuna.export.messages.NetworkMessage` with the identifier
        ``train_accuracy`` is published. '''
        Y           = message.data['network_output']
        labels      = message.data['input_labels']
        predictions = Y.argmax(1)
        n_correct   = (predictions == labels).sum().item()
        accuracy    = n_correct / float(labels.numel())
        self._backend.add_data('train_batch_accuracy', accuracy, message.global_step)

        kind = 'train_accuracy'
        self.message_bus.publish_network_message(message.global_step,
                                                 message.train_step,
                                                 message.epoch, kind,
                                                 data=accuracy)



from .utils import *
from .module_tree import ModuleTree
from .named_module import NamedModule

__all__ = ['ModuleTree', 'NamedModule', 'make_fill_polygons', 'available_optimizers',
           'create_optimizer', 'initialize_model', 'load_dataset']



'''This module contains functionality for numba-torch interoperability. This isn't used in the
library but may be useful in the future. Documentation is spotty.'''
import numba
import numpy as np
import numba.cuda as cuda
import sys
import torch


@numba.jit(nopython=True)
def compute_bin(x, n, xmin, xmax):
    '''Compute the bins for histogram computation'''
    # special case to mirror NumPy behavior for last bin
    if x == xmax:
        return n - 1    # a_max always in last bin

    bin = np.int32(n * (x - xmin) / (xmax - xmin))

    if bin < 0 or bin >= n:
        return None
    else:
        return bin


@cuda.jit
def histogram(x, xmin, xmax, histogram_out):
    '''Compute a histogram on a numba device array.'''
    nbins = histogram_out.shape[0]

    start = cuda.grid(1)
    stride = cuda.gridsize(1)

    for i in range(start, x.shape[0], stride):
        # note that calling a numba.jit function from CUDA automatically
        # compiles an equivalent CUDA device function!
        bin_number = compute_bin(x[i], nbins, xmin, xmax)

        if bin_number >= 0 and bin_number < histogram_out.shape[0]:
            cuda.atomic.add(histogram_out, bin_number, 1)


@cuda.jit
def min_max(x, min_max_array):
    '''Compute minimum and maximum in parallel on the gpu'''
    start = cuda.grid(1)
    stride = cuda.gridsize(1)

    # Array already seeded with starting values appropriate for x's dtype
    # Not a problem if this array has already been updated
    local_min = min_max_array[0]
    local_max = min_max_array[1]

    for i in range(start, x.shape[0], stride):
        element = x[i]
        local_min = min(element, local_min)
        local_max = max(element, local_max)

    # Now combine each thread local min and max
    cuda.atomic.min(min_max_array, 0, local_min)
    cuda.atomic.max(min_max_array, 1, local_max)


def dtype_min_max(dtype):
    '''Get the min and max value for a numeric dtype'''
    if np.issubdtype(dtype, np.integer):
        info = np.iinfo(dtype)
    else:
        info = np.finfo(dtype)
    return info.min, info.max


def numba_gpu_histogram(a_gpu, bins):
    '''Compute a histogram on a numba device array.'''
    # Find min and max value in array
    dtype_min, dtype_max = dtype_min_max(a_gpu.dtype)
    # Put them in the array in reverse order so that they will be replaced by
    # the first element in the array
    min_max_array_gpu = cuda.to_device(np.array([dtype_max, dtype_min], dtype=a_gpu.dtype))
    min_max[64, 64](a_gpu, min_max_array_gpu)
    a_min, a_max = min_max_array_gpu.copy_to_host()

    # Bin the data into a histogram
    histogram_out = cuda.to_device(np.zeros(shape=(bins,), dtype=np.int32))
    histogram[64, 64](a_gpu, a_min, a_max, histogram_out)

    return histogram_out


##################
#  Tensor2Numba  #
##################
def typestr(tensor):
    endianness = '<' if sys.byteorder == 'little' else '>'
    types = {
        torch.float32: 'f4',
        torch.float: 'f4',
        torch.float64: 'f8',
        torch.double: 'f8',
        torch.float16: 'f2',
        torch.half: 'f2',
        torch.uint8: 'u1',
        torch.int8: 'i1',
        torch.int16: 'i2',
        torch.short: 'i2',
        torch.int32: 'i4',
        torch.int: 'i4',
        torch.int64: 'i8',
        torch.long: 'i8'
    }
    return endianness + types[tensor.dtype]


def tensor_to_numba(tensor):
    '''Convert a tensor to a numba device array. The tensor will share the underlying CUDA storage
    This works by setting the ``__cuda_array_interface__`` attribute on the tensor.

    Parameters
    ----------
    tensor  :   torch.Tensor
    '''
    cai_dict = {
        'shape': tuple(tensor.shape),
        'data': (tensor.data_ptr(), True),
        'typestr': typestr(tensor),
        'version': 0,
        'strides': list(s * tensor.storage().element_size() for s in tensor.stride()),
        'descr': [('', typestr(tensor))]
    }
    setattr(tensor, '__cuda_array_interface__', cai_dict)
    device_array = cuda.as_cuda_array(tensor)
    return device_array



'''From https://github.com/gpleiss/efficient_densenet_pytorch/blob/master/models/densenet.py'''
# This implementation is based on the DenseNet-BC implementation in torchvision
# https://github.com/pytorch/vision/blob/master/torchvision/models/densenet.py

import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as cp
from collections import OrderedDict


def _bn_function_factory(norm, relu, conv):
    def bn_function(*inputs):
        concated_features = torch.cat(inputs, 1)
        bottleneck_output = conv(relu(norm(concated_features)))
        return bottleneck_output

    return bn_function


class _DenseLayer(nn.Module):
    def __init__(self, num_input_features, growth_rate, bn_size, drop_rate, efficient=False):
        super(_DenseLayer, self).__init__()
        self.add_module('norm1', nn.BatchNorm2d(num_input_features)),
        self.add_module('relu1', nn.ReLU(inplace=False)),
        self.add_module('conv1', nn.Conv2d(num_input_features, bn_size *
                        growth_rate, kernel_size=1, stride=1, bias=False)),
        self.add_module('norm2', nn.BatchNorm2d(bn_size * growth_rate)),
        self.add_module('relu2', nn.ReLU(inplace=False)),
        self.add_module('conv2', nn.Conv2d(bn_size * growth_rate, growth_rate,
                        kernel_size=3, stride=1, padding=1, bias=False)),
        self.drop_rate = drop_rate
        self.efficient = efficient

    def forward(self, *prev_features):
        bn_function = _bn_function_factory(self.norm1, self.relu1, self.conv1)
        if self.efficient and any(prev_feature.requires_grad for prev_feature in prev_features):
            bottleneck_output = cp.checkpoint(bn_function, *prev_features)
        else:
            bottleneck_output = bn_function(*prev_features)
        new_features = self.conv2(self.relu2(self.norm2(bottleneck_output)))
        if self.drop_rate > 0:
            new_features = F.dropout(new_features, p=self.drop_rate, training=self.training)
        return new_features


class _Transition(nn.Sequential):
    def __init__(self, num_input_features, num_output_features):
        super(_Transition, self).__init__()
        self.add_module('norm', nn.BatchNorm2d(num_input_features))
        self.add_module('relu', nn.ReLU(inplace=False))
        self.add_module('conv', nn.Conv2d(num_input_features, num_output_features,
                                          kernel_size=1, stride=1, bias=False))
        self.add_module('pool', nn.AvgPool2d(kernel_size=2, stride=2))


class _DenseBlock(nn.Module):
    def __init__(self, num_layers, num_input_features, bn_size, growth_rate, drop_rate,
                 efficient=False):
        super(_DenseBlock, self).__init__()
        for i in range(num_layers):
            layer = _DenseLayer(
                num_input_features + i * growth_rate,
                growth_rate=growth_rate,
                bn_size=bn_size,
                drop_rate=drop_rate,
                efficient=efficient,
            )
            self.add_module('denselayer%d' % (i + 1), layer)

    def forward(self, init_features):
        features = [init_features]
        for name, layer in self.named_children():
            new_features = layer(*features)
            features.append(new_features)
        return torch.cat(features, 1)


class DenseNet(nn.Module):
    r'''Densenet-BC model class, based on
    `"Densely Connected Convolutional Networks" <https://arxiv.org/pdf/1608.06993.pdf>`

    Parameters
    ----------
    growth_rate :   int
                    how many filters to add each layer (``k`` in paper)
    block_config    :   list(int)
                        3 or 4 ints,  how many layers in each pooling block
    num_init_features   :   int
                            the number of filters to learn in the first convolution layer
    bn_size :   int
                multiplicative factor for number of bottle neck layers (i.e. ``bn_size`` * ``k``
                features in the bottleneck layer)
    drop_rate   :   float
                    dropout rate after each dense layer
    num_classes :   int
                    number of classification classes
    small_inputs    :   bool
                        set to True if images are ``32x32`` (e.g. CIFAR). Otherwise assumes images
                        are larger.
    efficient   :   bool
                    set to ``True`` to use checkpointing. Much more memory efficient, but slower.
    exporter    :   ikkuna.export.Exporter
                    Optional exporter to use for monitoring
    '''
    def __init__(self, input_shape, growth_rate=12, block_config=(16, 16, 16), compression=0.5,
                 num_init_features=24, bn_size=4, drop_rate=0,
                 num_classes=10, small_inputs=True, efficient=False, exporter=None):

        super(DenseNet, self).__init__()
        assert 0 < compression <= 1, 'compression of densenet should be between 0 and 1'
        self.avgpool_size = 8 if small_inputs else 7

        _, _, C = input_shape

        # First convolution
        if small_inputs:
            self.features = nn.Sequential(OrderedDict([
                ('conv0', nn.Conv2d(C, num_init_features, kernel_size=3, stride=1, padding=1,
                                    bias=False)),
            ]))
        else:
            self.features = nn.Sequential(OrderedDict([
                ('conv0', nn.Conv2d(C, num_init_features, kernel_size=7, stride=2, padding=3,
                                    bias=False)),
            ]))
            self.features.add_module('norm0', nn.BatchNorm2d(num_init_features))
            self.features.add_module('relu0', nn.ReLU(inplace=False))
            self.features.add_module('pool0', nn.MaxPool2d(kernel_size=3, stride=2, padding=1,
                                                           ceil_mode=False))

        # Each denseblock
        num_features = num_init_features
        for i, num_layers in enumerate(block_config):
            block = _DenseBlock(
                num_layers=num_layers,
                num_input_features=num_features,
                bn_size=bn_size,
                growth_rate=growth_rate,
                drop_rate=drop_rate,
                efficient=efficient,
            )
            self.features.add_module('denseblock%d' % (i + 1), block)
            num_features = num_features + num_layers * growth_rate
            if i != len(block_config) - 1:
                trans = _Transition(num_input_features=num_features,
                                    num_output_features=int(num_features * compression))
                self.features.add_module('transition%d' % (i + 1), trans)
                num_features = int(num_features * compression)

        # Final batch norm
        self.features.add_module('norm_final', nn.BatchNorm2d(num_features))

        # Linear layer
        self.classifier = nn.Linear(num_features, num_classes)

        # Initialization
        for name, param in self.named_parameters():
            if 'conv' in name and 'weight' in name:
                n = param.size(0) * param.size(2) * param.size(3)
                param.data.normal_().mul_(math.sqrt(2. / n))
            elif 'norm' in name and 'weight' in name:
                param.data.fill_(1)
            elif 'norm' in name and 'bias' in name:
                param.data.fill_(0)
            elif 'classifier' in name and 'bias' in name:
                param.data.fill_(0)

        if exporter:
            exporter.set_model(self)
            exporter.add_modules(self)

    def forward(self, x):
        features = self.features(x)
        out = F.relu(features, inplace=False)
        out = F.avg_pool2d(out, kernel_size=self.avgpool_size).view(features.size(0), -1)
        out = self.classifier(out)
        return out



from .backend import TBBackend, MPLBackend, Backend, NullBackend, configure_prefix, set_run_info

__all__ = ['TBBackend', 'MPLBackend', 'Backend', 'configure_prefix', 'NullBackend']

backend_choices = ('tb', 'mpl')


def get_backend(name, plot_config, **kwargs):
    if name is None:
        return NullBackend(**plot_config, **kwargs)
    if name not in backend_choices:
        raise ValueError(f'Backend must be one of {backend_choices}')
    if name == 'tb':
        return TBBackend(**plot_config, **kwargs)
    if name == 'mpl':
        return MPLBackend(**plot_config, **kwargs)



#!/usr/bin/env python

from distutils.core import setup
import setuptools
import os
import ikkuna
import warnings

# read the requirements from requirements.txt. Some specs need special handing as `install_requires`
# understands only a subset of what `pip install` understands
with open('requirements.txt', mode='r') as f:
    requirements = set(f.readlines())
    pip_pkgs     = set(r for r in requirements if not '@' in r)
    repositories = requirements.difference(pip_pkgs)

# we use the readme as the description to display on pypi
with open('README.md') as f:
    readme = f.read()

# these are the entry_points which always work
entry_points = {
          'ikkuna.export.subscriber': [
              'HistogramSubscriber = ikkuna.export.subscriber.histogram:HistogramSubscriber',
              'RatioSubscriber = ikkuna.export.subscriber.ratio:RatioSubscriber',
              'SpectralNormSubscriber = ikkuna.export.subscriber.spectral_norm:SpectralNormSubscriber',
              'TestAccuracySubscriber = ikkuna.export.subscriber.test_accuracy:TestAccuracySubscriber',
              'TrainAccuracySubscriber = ikkuna.export.subscriber.train_accuracy:TrainAccuracySubscriber',
              'VarianceSubscriber = ikkuna.export.subscriber.variance:VarianceSubscriber',
              'NormSubscriber = ikkuna.export.subscriber.norm:NormSubscriber',
              'MeanSubscriber = ikkuna.export.subscriber.mean:MeanSubscriber',
              'MessageMeanSubscriber = ikkuna.export.subscriber.message_mean:MessageMeanSubscriber',
              'LossSubscriber = ikkuna.export.subscriber.loss:LossSubscriber',
          ]
      }

# if hessian_eigenthings is installed, we can supply the HessianEigenSubscriber, otherwise simply
# drop it
try:
    import hessian_eigenthings  # noqa
    entry_points['ikkuna.export.subscriber'].append(
        'HessianEigenSubscriber = ikkuna.export.subscriber.hessian_eig:HessianEigenSubscriber'
    )
except ImportError:
    warnings.warn('pytorch-hessian-eigenthings could not be imported. '
                  '`HessianEigenSubscriber` will not be installed. You can find the package at '
                  '`https://github.com/noahgolmant/pytorch-hessian-eigenthings/`')

# Problem: PyPi does not allow the syntax introduced in Pep 508 for listing requirements from a vcs
# repository, although setuptools understands it. This is supposedly because the developers want to
# avoid PyPi packages pulling from arbitrary URLs. So, as a workaround, we manually pull from this
# arbitrary url and install the SVCCA package. In the future, SVCCA should be on PyPi as well,
# making this unnecessary, but for now it's not tested enough.
try:
    import svcca  # noqa
    entry_points['ikkuna.export.subscriber'].extend(
        [
            'SVCCASubscriber = ikkuna.export.subscriber.svcca:SVCCASubscriber',
            'BatchedSVCCASubscriber = ikkuna.export.subscriber.svcca:BatchedSVCCASubscriber',
        ]
    )
except ImportError:
    cwd = os.getcwd()
    svcca_url = 'https://github.com/themightyoarfish/svcca-gpu/archive/master.zip'
    archive_dirname = 'svcca-gpu-master'
    warnings.warn(f'SVCCA is not installed. Installing from `{svcca_url}`')
    import tempfile
    import importlib
    from urllib.request import urlretrieve
    from zipfile import ZipFile
    with tempfile.TemporaryDirectory() as tempdir:
        os.chdir(tempdir)
        print(f'Downloading `{svcca_url}`')
        fname, headers = urlretrieve(svcca_url, f'{tempdir}/svcca.zip')
        f = ZipFile(fname)
        extracted_repo_path = os.path.join(tempdir, archive_dirname)
        print(f'Extracting `{archive_dirname}`')
        f.extractall()
        os.chdir(extracted_repo_path)
        print('Running setup script.')
        module_spec = importlib.util.spec_from_file_location('setup',
                                                             os.path.join(extracted_repo_path,
                                                                          'setup.py'))
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        os.chdir(cwd)
        print('Done.')

setup(name='ikkuna',
      version=ikkuna.__version__,
      description='Ikkuna Neural Network Monitor',
      long_description=readme,
      long_description_content_type='text/markdown',
      author='Rasmus Diederichsen',
      author_email='rasmus@peltarion.com',
      url='https://peltarion.github.io/ai_ikkuna/',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering :: Visualization',
                   'Topic :: Scientific/Engineering :: Artificial Intelligence',
                   'Programming Language :: Python :: 3.6',
                   'License :: OSI Approved :: MIT License',
                   'Intended Audience :: Developers',
                   ],
      keywords='deep-learning pytorch neural-networks machine-learning'.split(),
      packages=setuptools.find_packages('.', include=['ikkuna.*']),
      install_requires=list(pip_pkgs),
      entry_points=entry_points,
      zip_safe=False,   # don't install egg, but source
      )



from collections import defaultdict
from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class MessageMeanSubscriber(PlotSubscriber):
    '''Compute the mean over all messages with scalar data of a given kind in one train step. This
    is useful for displaying summary statistics.'''

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):

        if not isinstance(kind, str):
            raise ValueError('MessageMeanSubscriber only accepts 1 kind')

        title        = f'message_means'
        ylabel       = 'Mean'
        xlabel       = 'Train step'
        subscription = Subscription(self, [kind, 'batch_finished'], tag=tag, subsample=subsample)
        super().__init__([subscription], message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)
        self._add_publication(f'{kind}_message_mean', type='META')
        # kind -> list
        self._buffer = defaultdict(list)

    def compute(self, message):
        '''A :class:`~ikkuna.export.messages.ModuleMessage` with the identifier
        ``{kind}_message_mean`` is published. '''

        if message.kind == 'batch_finished':
            for kind, values in self._buffer.items():
                mean = sum(values) / len(values)
                topic = f'{kind}_message_mean'
                self._backend.add_data(topic, mean, message.global_step)

                self.message_bus.publish_network_message(message.global_step,
                                                         message.train_step,
                                                         message.epoch, topic, data=mean)
                self._buffer[kind] = []
        else:
            self._buffer[message.kind].append(message.data)



'''
.. module:: module_tree

.. moduleauthor:: Rasmus Diederichsen

This module defines the :class:`~ikkuna.utils.ModuleTree` class for easily traversing a module
hierarchy in order to generate unique hierarchical names for all involved modules to be used as
dictinary keys.
'''
import re
from collections import defaultdict

NUMBER_REGEX = re.compile(r'\d+')


class ModuleTree(object):
    '''
    Attributes
    ----------
    _module :   torch.nn.Module
    _name   :   str
                Hierarchical name for this module
    _children   :   list(ModuleTree)
                    Children of this module. Can be empty.
    _type_counter   :   dict(int)
                        Dict for keeping track of number of child modules of each class. Used for
                        disambiguating e.g. different successive conv layers which are children of
                        the same sequential module.
    '''

    def __init__(self, module, name=None, drop_name=True, recursive=True):
        '''
        Parameters
        ----------
        module  :   torch.nn.Module
        name    :   str or None
                    If no name is given, one will be generated. If ``drop_name == True``, this
                    parameter is ignored
        drop_name   :   bool
                        Ignore the given name and set it to ``''``. Useful for dropping the root
                        name (e.g. ``alexnet``) lest it appear in every child name.
        recursive   :   bool
                        Add all :meth:`torch.nn.Module.named_children` as children to this tree
        '''
        # for the root node, it often makes sense not to include the name, since it will not add any
        # information
        if drop_name:
            name = ''
        else:
            # if the name is just the index, assume it's autogenerated by e.g. nn.Sequential and
            # make a better one
            if name is None or re.match(NUMBER_REGEX, name):
                name = module.__class__.__name__.lower()

        self._module       = module
        self._name         = name
        self._children     = []
        self._type_counter = defaultdict(int)

        if recursive:
            named_children = list(module.named_children())

            # children could be empty
            if named_children:
                for child_name, child in named_children:
                    child_class = child.__class__
                    # again, if it's just an index, make a new name. TODO: Figure out how to
                    # deduplicate this
                    if re.match(NUMBER_REGEX, child_name):
                        class_name = child_class.__name__.lower()
                        class_index = self._type_counter[child_class]
                        child_name = f'{class_name}{class_index}'

                    self._children.append(ModuleTree(child,
                                                     name=f'{self._name}/{child_name}',
                                                     drop_name=False,
                                                     recursive=True)
                                          )
                    self._type_counter[child.__class__] += 1

    def preorder(self, depth=-1):
        '''Traverse the tree in preorder.

        Parameters
        ----------
        depth   :   int
                    Depth to which to traverse the tree. -1 means infinite depth.

        Yields
        ------
        ikkuna.utils.NamedModule
            Tuple of name, module and parent module
        '''
        # due to some weird bug, we cannot import this at the top despite the fact that there is no
        # circular import
        from ikkuna.utils import NamedModule

        if depth == 0 or not self._children:
            yield NamedModule(name=self._name, module=self._module)
        elif depth > 0:
            depth -= 1
            for child in self._children:
                yield from child.preorder(depth)
        else:
            for child in self._children:
                yield from child.preorder(depth=-1)



from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class HistogramSubscriber(PlotSubscriber):

    '''A :class:`~ikkuna.export.subscriber.Subscriber` which subsamples training artifacts and
    computes histograms per epoch.  Histograms are non-normalized.
    '''

    def __init__(self, kind, message_bus=get_default_bus(), tag='default', subsample=1, backend='tb'):

        if not isinstance(kind, str):
            raise ValueError('HistogramSubscriber only accepts 1 kind')

        subscription = Subscription(self, [kind], tag=tag, subsample=subsample)
        title        = f'{kind}_histogram'
        ylabel       = 'Frequency'
        super().__init__([subscription], message_bus, {'title': title, 'ylabel': ylabel},
                         backend=backend)

    def compute(self, message):
        '''
        .. note::
            Since the histogram is computed by the visualization backend (there's no practical way
            around it), this subscriber does *not* publish a
            :class:`~ikkuna.export.messages.ModuleMessage`
        '''

        module, name = message.key
        data         = message.data
        self._backend.add_histogram(name, data, message.global_step)



'''

'''
import abc


META_KINDS = {
    'batch_started', 'batch_finished', 'epoch_started', 'epoch_finished', 'input_data', 'loss',
    'input_labels', 'network_output'
}
'''Message kinds which are not tied to any specific module. These topics is just what comes with
the library, others can be added to a specific :class:`MessageBus`'''

DATA_KINDS = {
    'weights', 'weight_gradients', 'weight_updates', 'biases', 'bias_gradients', 'bias_updates',
    'activations', 'layer_gradients'
}
'''Message kinds which are tied to a specific module and always carry data. These topics is just
what comes with the library, others can be added to a specific :class:`MessageBus`'''


class Message(abc.ABC):
    '''Base class for messages emitted from the :class:`~ikkuna.export.Exporter`.

    These messages are assembled into :class:`MessageBundle` objects in the
    :class:`~ikkuna.export.subscriber.Subscription`.
    '''
    def __init__(self, tag, global_step, train_step, epoch, kind):
        '''
        Parameters
        ----------
        tag :   str
                Tag for this message
        global_step :   int
                Global train step
        train_step    :   int
                    Epoch-local train step
        epoch   :   int
                    Epoch index
        kind    :   str
                    Message topic
        '''
        self._tag = tag
        self._global_step = global_step

        # check train_step
        if train_step < 0:
            raise ValueError('Step cannot be negative.')
        else:
            self._train_step = train_step

        # check epoch
        if epoch < 0:
            raise ValueError('Epoch cannot be negative')
        else:
            self._epoch = epoch

        self._kind = kind

        self._data = None

    @property
    def tag(self):
        ''' str: The tag associated with this message '''
        return self._tag

    @property
    def global_step(self):
        '''int: Global sequence number. This counter should not reset after each epoch.'''
        return self._global_step

    @property
    def train_step(self):
        '''int: Epoch-local sequence number (the current batch index)'''
        return self._train_step

    @property
    def epoch(self):
        '''int: Current epoch number'''
        return self._epoch

    @property
    def kind(self):
        '''str: Message kind'''
        return self._kind

    @property
    def data(self):
        '''torch.Tensor, tuple(torch.Tensor) or None:  This field is optional for
        :class:`NetworkMessage`, but mandatory for :class:`ModuleMessage`'''
        return self._data

    @abc.abstractproperty
    def key(self):
        '''object: A key used for grouping messages into :class:`MessageBundle` s'''
        pass

    def __str__(self):
        return (f'<{self.__class__.__name__}: global_step={self.global_step}, '
                f'train_step={self.train_step}, epoch={self.epoch}, kind={self.kind}>')

    def __repr__(self):
        return str(self)


class NetworkMessage(Message):
    '''A message with meta information not tied to any specific module. Can still carry tensor data,
    if necessary.'''
    def __init__(self, tag, global_step, train_step, epoch, kind, data=None):
        super().__init__(tag, global_step, train_step, epoch, kind)
        self._data = data

    @property
    def data(self):
        '''torch.Tensor, tuple, float, int or None: Optional data. Can be used e.g. for input to the
        network, labels or network output'''
        return self._data

    @property
    def key(self):
        return 'META'


class ModuleMessage(Message):
    '''A message tied to a specific module, with tensor data attached.'''

    def __init__(self, tag, global_step, train_step, epoch, kind, named_module, data):
        super().__init__(tag, global_step, train_step, epoch, kind)
        self._module  = named_module
        if data is None:
            raise ValueError('Data cannot be `None` for `ModuleMessage`')
        self._data = data

    @property
    def module(self):
        '''torch.nn.Module: Module emitting this data'''
        return self._module

    @property
    def key(self):
        return self.module


class MessageBundle(object):

    '''Data object for holding a set of artifacts for a module (or meta information) at one point
    during training. This data type can be used to buffer different kinds and check whether all
    expected kinds have been received for a module or meta information. The collection is enforced
    to be homogeneous with respect to global step, train step, epoch, and identifier
    (:attr:`Message.key`)'''

    def __init__(self, kinds):
        '''
        Parameters
        ----------
        kinds   :   str or list
                    Single kind when a :class:`~ikkuna.export.subscriber.Subscription` is used, or a
                    list of Message kinds contained in this bundle for use with
                    :class:`~ikkuna.export.subscriber.SynchronizedSubscription`
        '''
        if isinstance(kinds, str):
            # this has bitten me before. base `Subscription`s don't use multiple kinds
            kinds = [kinds]
        self._key            = None
        self._expected_kinds = kinds
        self._received       = {kind: False for kind in kinds}
        self._data           = {kind: None for kind in kinds}
        self._global_step    = None
        self._train_step     = None
        self._epoch          = None

    @property
    def key(self):
        ''' str: An object denoting the common aspect of the collected messages (besides the steps).
        This can be the :class:`~ikkuna.utils.NamedModule` emitting the data or a string such as
        ``'META'`` or other denoting these are messages which do not belong to a module.'''
        return self._key

    @property
    def kinds(self):
        '''list(str): Alias to :attr:`expected_kinds`'''
        return self.expected_kinds

    @property
    def expected_kinds(self):
        '''list(str): The expected kinds of messages per iteration '''
        return self._expected_kinds

    @property
    def data(self):
        '''dict(str, torch.Tensor): The tensors received for each kind'''
        return self._data

    @property
    def global_step(self):
        '''int: Global sequence number of this class'''
        return self._global_step

    @property
    def train_step(self):
        '''int: Sequence number (training step) of the received messages (should match across all
        msgs in one iteration)'''
        return self._train_step

    @property
    def epoch(self):
        '''int: Epoch of the received messages (should match across all msgs in one iteration)'''
        return self._epoch

    def complete(self):
        '''Check if all expected messages have been received. This means the bundle can be released
        to subscribers.

        Returns
        -------
        bool
        '''
        return all(self._received.values())

    def check_message(self, message):
        '''Check consistency of sequence number, step and epoch or set if not set yet. Check
        consistency of identifier and check for duplication.

        Parameters
        ----------
        message :   ikkuna.export.messages.Message

        Raises
        ------
        ValueError
            If ``message.(global_step|step|epoch|identifier)`` does not match the current
            ``(global_step|step|epoch|identifier)`` or in case a message of ``message.kind`` has
            already been received
        '''
        #################
        #  global_step  #
        #################
        if self._global_step is None:
            self._global_step = message.global_step
        elif message.global_step != self._global_step:
            raise ValueError('Attempting to add message with global_step '
                             f'{message.global_step} to bundle with '
                             'initial global_step {self._global_step}')

        ################
        #  train_step  #
        ################
        if self._train_step is None:
            self._train_step = message.train_step
        elif message.train_step != self._train_step:
            raise ValueError(f'Attempting to add message with step {message.step} to bundle with '
                             'initial step {self._train_step}')
        #############
        #  epoch    #
        #############
        if self._epoch is None:
            self._epoch = message.epoch
        elif message.epoch != self._epoch:
            raise ValueError(f'Attempting to add message from epoch {message.epoch} to bundle with '
                             'initial epoch {self._epoch}')

        ################
        #  key         #
        ################
        if self._key is None:
            self._key = message.key
        elif self._key != message.key:
            raise ValueError(f'Unexpected message key "{message.key}" '
                             f'(expected "{self._key}")')

        #################
        #  Duplication  #
        #################
        if self._received[message.kind]:
            raise ValueError(f'Got duplicate value for kind "{message.kind}".')

    def add_message(self, message):
        '''Add a new message to this object. Will fail if the new messsage does not have the same
        sequence number and epoch.

        Parameters
        ----------
        message :   ikkuna.export.messages.Message

        Raises
        ------
        ValueError
            see :meth:`~MessageBundle.check_message()`
        '''
        self.check_message(message)
        self._received[message.kind] = True

        if message.data is not None:
            self._data[message.kind] = message.data

    def __getattr__(self, name):
        '''Override to mimick a property for each kind of message in this data (e.g.
        ``message_bundle.activations`` instead of ``message_bundle.data['activations']``)
        '''
        if name in self._expected_kinds:
            return self._data[name]
        else:
            return self.__getattribute__(name)

    def __str__(self):
        mod   = self._key
        step  = self._train_step
        kinds = list(self._data.keys())
        return f'<MessageBundle: identifier={mod}, kinds={kinds}, step={step}>'

    def __repr__(self):
        return str(self)


class MessageBus(object):
    '''A class which receives messages, registers subscribers and relays the former to the
    latter.'''

    def __init__(self, name):
        '''
        Parameters
        ----------
        name    :   str
                    Identifier for this bus
        '''
        self._name = name
        self._subscribers = set()
        self._meta_kinds = META_KINDS
        self._data_kinds = DATA_KINDS

    def register_meta_topic(self, kind):
        '''Register a topic so it can be subscribed.'''
        print(f'Registering topic {kind}')
        self._meta_kinds.add(kind)

    def deregister_meta_topic(self, kind):
        '''Unregister a topic so it can not be subscribed any longer.'''
        self._meta_kinds.remove(kind)
        # TODO: Don't just believe the sender that it registered the topic

    def register_data_topic(self, kind):
        '''Register a topic so it can be subscribed.'''
        print(f'Registering topic {kind}')
        self._data_kinds.add(kind)

    def deregister_data_topic(self, kind):
        '''Unregister a topic so it can not be subscribed any longer.'''
        self._data_kinds.remove(kind)
        # TODO: Don't just believe the sender that it registered the topic

    @property
    def name(self):
        '''str: The name of this bus'''
        return self._name

    def register_subscriber(self, sub):
        '''Add a new subscriber to the set. Adding subscribers mutliple times will still only call
        them once per message.

        Parameters
        ----------
        sub :   ikkuna.export.subscriber.Subscriber

        Raises
        ------
        ValueError
            If any of the kinds the Subscriber is interested in wasn't previously registered
        '''
        known_kinds = self._meta_kinds.union(self._data_kinds)
        for kind in sub.kinds:
            if kind not in known_kinds:
                raise ValueError(f'"{kind}" was not registered.')
        self._subscribers.add(sub)

    def publish_network_message(self, global_step, train_step, epoch, kind, data=None,
                                tag='default'):
        '''Publish an update of type :class:`~ikkuna.export.messages.NetworkMessage` to all
        registered subscribers.

        Parameters
        ----------
        global_step :   int
                        Global training step
        train_step  :   int
                        Epoch-relative training step
        epoch   :   int
                    Epoch index
        kind    :   str
                    Kind of message
        data    :   torch.Tensor or None
                    Payload, if necessary
        '''
        if kind not in self._meta_kinds:
            raise ValueError(f'Unknown META kind "{kind}". '
                             'Check spelling and kind of your publications.')

        msg = NetworkMessage(global_step=global_step, tag=tag, kind=kind, train_step=train_step,
                             epoch=epoch, data=data)
        for sub in self._subscribers:
            sub.receive_message(msg)

    def publish_module_message(self, global_step, train_step, epoch, kind, named_module, data,
                               tag='default'):
        '''Publish an update of type :class:`~ikkuna.export.messages.ModuleMessage` to all
        registered subscribers.

        Parameters
        ----------
        global_step :   int
                        Global training step
        train_step  :   int
                        Epoch-relative training step
        epoch   :   int
                    Epoch index
        kind    :   str
                    Kind of message
        named_module  :   ikkuna.utils.NamedModule
                    The module in question
        data    :   torch.Tensor
                    Payload
        '''
        if kind not in self._data_kinds:
            raise ValueError(f'Unknown DATA kind "{kind}". '
                             'Check spelling and kind of your publications.')
        msg = ModuleMessage(global_step=global_step, tag=tag, kind=kind, named_module=named_module,
                            train_step=train_step, epoch=epoch, data=data)

        for sub in self._subscribers:
            sub.receive_message(msg)


__default_bus = MessageBus('default')


def get_default_bus():
    '''Get the default message bus which is always created when this module is loaded.

    Returns
    -------
    MessageBus
    '''
    global __default_bus
    return __default_bus
